from proper import Controller
from proper.concerns import (
    RequestForgeryProtection,
    RestoreSession,
    UpdateSessionCookie,
)

from .concerns.security_headers import SetSecurityHeaders


class AppController(Controller):
    """All other controllers must inherit from this class.
    """
    # Note: The order might matter
    before = [
        RestoreSession(),
        RequestForgeryProtection(),
    ]
    after = [
        UpdateSessionCookie(),
        SetSecurityHeaders(),
    ]



import os

from proper.units import MB


env = os.getenv("APP_ENV", "dev")

DEBUG = env == "dev"

if env == "prod":
    PROTOCOL = "https"
    HOST = "YOUR-DOMAIN.com"
else:
    PROTOCOL = "http"
    HOST = "localhost:2300"

# List of secret keys, **oldest to newest**.
# Every key in the list is valid, so you can periodically generate a new key
# and remove the oldest one to add and extra layer of mitigation
# against an attacker discovering a secret key.
if env == "prod":
    SECRET_KEYS = os.getenv("SECRET_KEYS", "").split(",")
else:
    SECRET_KEYS = [
        "---- This is a not-secret-secret_key just for development ----"
    ]

# Turn off to let debugging WSGI middleware handle exceptions.
CATCH_ALL_ERRORS = True

# Limits the total content length (in bytes).
# Raises a RequestEntityTooLarge exception if this value is exceeded.
MAX_CONTENT_LENGTH = 8 * MB

# Limits the content length (in bytes) of the query string.
# Raises a RequestEntityTooLarge or an UriTooLong if this value is exceeded.
MAX_QUERY_SIZE = 1 * MB

STATIC_URL = "/static/"
VIEWS_ASSETS_URL = "/assets/"

# The name of the header to use to return a file
# so the proxy or web-server does it instead of our application.
# Lighttpd uses "X-Sendfile" while NGINX uses "X-/Accel-Redirect"
if env == "prod":
    STATIC_X_SENDFILE_HEADER = "X-Accel-Redirect"
else:
    STATIC_X_SENDFILE_HEADER = ""


MAILER = {
    "type": "proper.mail.ToConsoleMailer",
    "default_from": "hello@example.com",
}

if env == "test":
    MAILER["type"] = "proper.mail.ToMemoryMailer"

# if env == "prod":
#     MAILER = {
#         "type": "proper.mail.SMTPMailer",
#         "host": "smtp.example.com",
#         "port": 587,
#         "username": os.getenv("SMTP_USERNAME"),
#         "password": os.getenv("SMTP_PASSWORD"),
#         "use_tls": True,
#         "default_from": MAILER["default_from"],
#     }

update(storage.config)
update(session.config)



import peewee as pw

from app.main import app


db = app.db["main"]


class BaseModel(pw.Model):
    class Meta:
        database = app.db["main"]


class BaseMixin(pw.Model):
    class Meta:
        database = app.db["main"]



"""Peewee migrations -- 001_base.py.

Create the tables needed for Proper's SQL-based queues.
Delete this file if you are going to use a different queue backend.
"""

import peewee as pw
from peewee_migrate import Migrator


def migrate(migrator: Migrator, database: pw.Database, *, fake=False):
    @migrator.create_model
    class QueueKV(pw.Model):
        queue = pw.CharField(max_length=255)
        key = pw.CharField(max_length=255)
        value = pw.BlobField()

        class Meta:
            table_name = "queuekv"
            primary_key = pw.CompositeKey("queue", "key")

    @migrator.create_model
    class QueueSchedule(pw.Model):
        id = pw.AutoField()
        queue = pw.CharField(max_length=255)
        data = pw.BlobField()
        timestamp = pw.TimestampField()

        class Meta:
            table_name = "queueschedule"
            indexes = [(("queue", "timestamp"), False)]

    @migrator.create_model
    class QueueTask(pw.Model):
        id = pw.AutoField()
        queue = pw.CharField(max_length=255)
        uuid = pw.UUIDField(index=True)
        data = pw.BlobField()
        priority = pw.FloatField(default=0.0)
        done = pw.BooleanField(default=False)

        class Meta:
            table_name = "queuetask"
            indexes = [(("done", "priority", "id"), False)]


def rollback(migrator: Migrator, database: pw.Database, *, fake=False):
    migrator.remove_model("queuetask")
    migrator.remove_model("queueschedule")
    migrator.remove_model("queuekv")



from huey.consumer import Consumer
from app.main import app


def get_config():
    return app.config.get("QUEUE_CONSUMER", {}).copy()


def run_consumer(config):
    if app.queue is None:
        raise RuntimeError("Queue not initialized.")
    print("Starting background workers...")
    consumer = Consumer(app.queue, **options)
    consumer.run()


def run_consumer_proc():
    import multiprocessing
    import sys

    if sys.platform == "darwin":
        try:
            multiprocessing.set_start_method("fork")
        except RuntimeError:
            pass

    config = get_config()
    consumer_proc = multiprocessing.Process(
        target=run_consumer,
        kwargs={"config": config},
    )
    consumer_proc.start()


if __name__ == "__main__":
    config = get_config()
    run_consumer(config)



import os

from proper.units import DAYS


env = os.getenv("APP_ENV", "dev")

SESSION_LIFETIME = 30 * DAYS  # Seconds to expire an unused session.
SESSION_COOKIE_NAME = "_session"
SESSION_COOKIE_DOMAIN = None
SESSION_COOKIE_PATH = "/"
SESSION_COOKIE_HTTPONLY = True
SESSION_COOKIE_SECURE = (env == "prod")
SESSION_COOKIE_SAMESITE = "Lax"



from proper.mail import ToConsoleMailer

from app.main import config


mailer = ToConsoleMailer()


def send_email(to, subject, **kw):
    kw.setdefault("from_email", config.MAILER_DEFAULT_FROM)
    mailer.send(to=to, subject=subject, **kw)



import peewee as pw

from app.models.base import BaseMixin


class Timestamped(BaseMixin):
    created_at = pw.DateTimeField(default=pw.utcnow, null=True)
    updated_at = pw.DateTimeField(default=pw.utcnow, null=True)

    @classmethod
    def update(cls, *args, **kwargs):
        kwargs["updated_at"] = pw.utcnow()
        return super().update(*args, **kwargs)

    def save(self, *args, **kwargs):
        self.updated_at = pw.utcnow()
        return super().save(*args, **kwargs)



from proper import errors

from .app import AppController
from app.router import router


class PublicController(AppController):
    # Uncomment to add a home page
    # @router.get("")
    # def index(self):
    #     return self.render("public.index")

    @router.error(errors.NotFound)
    @router.get("_not_found")
    def not_found(self):
        return self.render("public.not-found")

    @router.error(Exception)
    @router.get("_error")
    def error(self):
        return self.render("public.error")



"""
Import here all the models you want to be detected for migrations.
You can also do `from .your_model import *  #noqa` here
and, in your model file, use `__all__` to define what to import

Do not import `BaseModel` or `BaseMixin` here.
"""
from .base import db  # noqa



import os
import typing as t


env = os.getenv("APP_ENV", "dev")

DATABASES: dict[str, t.Any] = {
    "main": {
        "type": "playhouse.sqlite_ext.SqliteExtDatabase",
        "database": "storage/app.sqlite3",
    },
    # "main": {
    #     "type": "playhouse.postgres_ext.PostgresqlExtDatabase",
    #     "database": os.getenv("DB_NAME", "[[app_name]]"),
    #     "host": os.getenv("DB_HOST", "127.0.0.1"),
    #     "port": int(os.getenv("DB_PORT", 5432)),
    #     "user": os.getenv("DB_USER", "root"),
    #     "password": os.getenv("DB_PASSWORD", ""),
    #     # The connection is managed in a concern of the controllers,
    #     # and on the `on_teardown` and `on_error` hooks
    #     "autoconnect": False,
    # }
}


QUEUE = {
    "type": "huey.SqliteHuey",
    "database": "storage/queue.sqlite3",
}
QUEUE_CONSUMER = {
    # Number of workers to spawn.
    "workers": 1,
    # Enable periodic task scheduler?
    "periodic": True,
    # Default queue polling interval.
    "initial_delay": 0.1,
    # Exponential backoff factor when queue empty.
    "backoff": 1.15,
    # Maximum interval between polling events.
    "max_delay": 10.0,
    # Interval for the scheduler. Must be between 1 and 60s
    "scheduler_interval": 1,
    # Type of worker to use ("thread", "process", or "greenlet").
    "worker_type": "thread",
    # Whether to check worker health.
    "check_worker_health": True,
    # Interval for health checks.
    "health_check_interval": 10,
    # Whether to flush locks.
    "flush_locks": False,
    # Comma-separated extra locks to use.
    "extra_locks": "",
}

CACHE = {
    "type": "proper.cache.SqliteCache",
    "database": ":memory:",
}


# --- Override config for testing ---
if env == "test":
    DATABASES["main"] = {
        "type": "playhouse.sqlite_ext.SqliteExtDatabase",
        "database": ":memory:",
    }

    QUEUE = {
        "type": "huey.MemoryHuey",
        "immediate": True,
        "immediate_use_memory": True,
    }

    CACHE = {
        "type": "proper.cache.NoCache",
    }


# --- Override config for production ---
if env == "prod":
    DATABASES["main"] = {
        "type": "playhouse.postgres_ext.PostgresqlExtDatabase",
        "database": os.getenv("DB_NAME", "[[app_name]]"),
        "host": os.getenv("DB_HOST", "127.0.0.1"),
        "port": int(os.getenv("DB_PORT", 5432)),
        "user": os.getenv("DB_USER", "root"),
        "password": os.getenv("DB_PASSWORD", ""),
        # The connection is managed in a concern of the controllers,
        # and on the `on_teardown` and `on_error` hooks
        "autoconnect": False,
    }

    QUEUE = {
        "type": "huey.contrib.sql_huey.SqlHuey",
        "database": os.getenv("DB_QUEUE_NAME", "[[app_name]]_queue"),
        "host": os.getenv("DB_QUEUE_HOST", "127.0.0.1"),
        "port": int(os.getenv("DB_QUEUE_PORT", 5432)),
        "user": os.getenv("DB_QUEUE_USER", "root"),
        "password": os.getenv("DB_QUEUE_PASSWORD", ""),
        # The connection is managed in a concern of the controllers,
        # and on the `on_teardown` and `on_error` hooks
        "autoconnect": False,
    }

    CACHE = {
        "type": "proper.cache.SqliteCache",
        "database": "storage/cache.sqlite3",
    }



from proper import show_banner, show_welcome


host = "0.0.0.0"
port = 2300


wsgi_app = "app.main:app"

# The socket to bind.
# A string of the form: 'HOST', 'HOST:PORT', 'unix:PATH'.
# An IP is a valid HOST.
bind = f"{host}:{port}"

reload = True
reload_extra_files = []


# The number of worker processes that this server
# should keep alive for handling requests.
#
# A positive integer generally in the 2-4 x $(NUM_CORES)
# range. You'll want to vary this a bit to find the best
# for your particular application's work load.
workers = 1

# The type of workers to use. The default
# sync class should handle most 'normal' types of work
# loads. You'll want to read
# http://docs.gunicorn.org/en/latest/design.html#choosing-a-worker-type
# for information on when you might want to choose one
# of the other worker classes.
worker_class = "sync"

# For the eventlet and gevent worker classes
# this limits the maximum number of simultaneous clients that
# a single process can handle.
#
# A positive integer generally set to around 1000.
worker_connections = 1000

# If a worker does not notify the master process in this
# number of seconds it is killed and a new worker is spawned
# to replace it.
#
# Generally set to thirty seconds. Only set this noticeably
# higher if you're sure of the repercussions for sync workers.
# For the non sync workers it just means that the worker
# process is still communicating and is not tied to the length
# of time required to handle a single request.
timeout = 30

# The number of seconds to wait for the next request
# on a Keep-Alive HTTP connection.
#
# A positive integer. Generally set in the 1-5 seconds range.
keepalive = 2


# The path to a log file to write to.
# A path string. "-" means log to stdout.
accesslog = "-"
errorlog = "-"

# The granularity of log output
# A string of "debug", "info", "warning", "error", "critical"
loglevel = "info"

access_log_format = "%(r)s | %(B)s bytes in %(M)s ms"


#A base to use with setproctitle to change the way
# that Gunicorn processes are reported in the system process
# table. This affects things like 'ps' and 'top'.
proc_name = "[[ app_name ]]"


def on_starting(server):
    show_banner()


def when_ready(server):
    show_welcome(host, port)


def pre_exec(server):
    show_welcome(host, port)



import argparse
import os
import sys
from pathlib import Path

import inflection
from hecto import (
    COLORS,
    printf,
    render_blueprint,
)


APP_BLUEPRINT = Path(__file__).parent / "blueprint"


def call(cmd: str) -> None:
    printf("run", cmd, color=COLORS.OK)
    os.system(cmd)


def gen_app(
    path: str | Path,
    *,
    name: str = "",
    force: bool = False,
    install_deps: bool = True,
) -> None:
    """Creates a new Proper application at `path`.

    Args:
        path:
            Where to create the new application.
        name:
            Optional name of the app instead of the one in `path`
        force:
            Overwrite files that already exist, without asking.
        install_deps:
            Whether to install dependencies after generating the app.

    """
    path = Path(path).resolve().absolute()
    path.mkdir(parents=True, exist_ok=False)
    app_name = inflection.underscore(name or str(path.stem))

    render_blueprint(
        APP_BLUEPRINT,
        path,
        context={
            "app_name": app_name,
        },
        force=force,
    )
    print()

    if install_deps:
        _install_dependencies(path)
    _wrap_up(path)


def _install_dependencies(path: Path) -> None:
    os.chdir(path)
    venv_path = str(path / ".venv")
    os.environ["VIRTUAL_ENV"] = venv_path
    call(f"""cd {str(path)} \\
            && uv venv \\
            && uv sync --group dev
    """)
    # call("tailwindcss_install")


def _wrap_up(path: Path) -> None:
    print("✨ Done! ✨")
    print()
    print(" The following steps are missing:")
    print()
    print("   $ cd " + path.stem + "")
    print("   $ source .venv/bin/activate")
    print()
    print(" Start your Proper app with:")
    print()
    print("   $ proper run")
    print()


def main():
    usage = "uvx proper-new <path> [--name <app_name>] [--force]"
    description="""
    The `proper-new` command creates a new Proper application at the path you specify.
    """.strip()

    if len(sys.argv) == 1:
        print("Usage:")
        print(f"    {usage}")
        sys.exit(1)

    parser = argparse.ArgumentParser(
        usage=usage,
        description=description
    )
    parser.add_argument("path", help="The required path argument")
    parser.add_argument("--name", help="Optional name of the app instead of the one in `path`", default="")
    parser.add_argument("--force", help="Overwrite files that already exist, without asking", action="store_true")
    args = parser.parse_args()
    gen_app(args.path, name=args.name, force=args.force)


if __name__ == "__main__":
    main()



"""
Routes without a controller.
Other routes are defined as decorators and mounted when
the controllers are imported.
"""
from .main import app


router = app.router

# Static files
router.static(app.config.STATIC_URL, root=app.static_path, name="static")

# Root-level static files
router.get("favicon.ico", redirect="/static/favicon.ico")
router.get("robots.txt", redirect="/static/robots.txt")
router.get("humans.txt", redirect="/static/humans.txt")



"""Auto-import all the views in this folder."""
from importlib import import_module
from inspect import isclass
from pathlib import Path
from pkgutil import iter_modules

from proper import Controller


classes = {}

# iterate through the modules in the current package
package_dir = str(Path(__file__).resolve().parent)
for (_, module_name, _) in iter_modules([package_dir]):
    # import the module and iterate through its attributes
    module = import_module(f"{__name__}.{module_name}")
    for attribute_name in dir(module):
        attribute = getattr(module, attribute_name)

        if isclass(attribute) and issubclass(attribute, Controller):
            classes[attribute_name] = attribute

globals().update(classes)



"""Peewee migrations -- 001_base.py.

Create the tables needed for Proper's database-based cache.
Delete this file if you are going to use a different cache backend.
"""

import peewee as pw
from peewee_migrate import Migrator


def migrate(migrator: Migrator, database: pw.Database, *, fake=False):
    @migrator.create_model
    class Cache(pw.Model):
        key = pw.TextField(primary_key=True)
        value = pw.BlobField()
        timestamp = pw.IntegerField(index=True)

        class Meta:
            table_name = "proper_cache"


def rollback(migrator: Migrator, database: pw.Database, *, fake=False):
    migrator.remove_model("proper_cache")



from proper import Controller


class SetSecurityHeaders:
    def __call__(self, co: Controller):
        # It determines if a web page can or cannot be included via <frame>
        # and <iframe> topics by untrusted domains.
        # https://developer.mozilla.org/Web/HTTP/Headers/X-Frame-Options
        co.response.headers.setdefault("X-Frame-Options", "SAMEORIGIN")

        # Determine the behavior of the browser in case an XSS attack is
        # detected. Use Content-Security-Policy without allowing unsafe-inline
        # scripts instead.
        # https://developer.mozilla.org/Web/HTTP/Headers/X-XSS-Protection
        co.response.headers.setdefault("X-XSS-Protection", "1", mode="block")

        # Download files or try to open them in the browser?
        co.response.headers.setdefault("X-Download-Options", "noopen")

        # Set to none to restrict Adobe Flash Player’s access to the web page data.
        co.response.headers.setdefault("X-Permitted-Cross-Domain-Policies", "none")

        co.response.headers.setdefault("Referrer-Policy", "strict-origin-when-cross-origin")



import pytest

from app.models import db
from app.models.base import BaseModel


@pytest.fixture(scope="session")
def db_setup():
    # better to be safe than sorry
    assert "_test" in db.database or db.database == ":memory:"
    models = BaseModel.__subclasses__()
    db.drop_tables(models)
    db.create_tables(models, safe=True)
    load_fixtures()
    yield
    db.drop_tables(models)


def load_fixtures():
    pass


@pytest.fixture(autouse=True)
def dbs(db_setup):
    with db.atomic() as transaction:
        yield
        transaction.rollback()


